"""matplatex: export matplotlib figures as pdf and text separately for
use in LaTeX.

Copyright (C) 2024 Johannes Sørby Heines

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from collections.abc import Iterator
from functools import cached_property
import sys

import matplotlib.pyplot as plt
from matplotlib.colors import to_rgba
from beartype import beartype

from .latex_input import LaTeXinput

def write_tex(
        output: LaTeXinput, fig, *,
        graphics, add_anchors=False, verbose=False
        ):
    output.includegraphics(graphics, get_height_to_width(fig))
    for element in extract_text(fig):
        if add_anchors:  # useful for checking positioning
            draw_anchors(fig, element.position_in_figure)
        if verbose:
            print(f"Adding {element}", end='\n', file=sys.stderr)
        output.add_text(
            element.text,
            position=element.position_in_figure,
            anchor=element.tikz_anchor,
            rotation=element.rotation,
            color=element.color)

class FigureText:
    """Contain text and its tikz properties."""
    def __init__(
            self,
            text: plt.Text,
            fig: plt.Figure,
            ax: plt.Axes | None):
        """Constructor for the FigureText class."""
        self.mpl_text = text
        self._figure_transform = fig.transFigure
        self._ax = ax

    def __str__(self):
        return (
            f"FigureText({self.mpl_text}, position={self.position_in_figure}, "
            f"visible={self.visible}, tikz_anchor={self.tikz_anchor})")

    @property
    def text(self) -> str:
        return self.mpl_text.get_text()

    @property  # Disconnect from attribute so I can change how it works.
    def position_in_figure(self) -> tuple[float]:
        return self._figure_xy

    @property
    def rotation(self) -> float:
        return self.mpl_text.get_rotation()

    @property
    def color(self) -> tuple[float]:
        return to_rgba(self.mpl_text.get_color())

    @color.setter
    def color(self, value, /):
        self.mpl_text.set_color(value)

    @cached_property
    def tikz_anchor(self) -> str:
        anchor_by_va = {
            'bottom': 'south',
            'top': 'north',
            'center': '',
            'baseline': 'base',
            'center_baseline': 'mid'
            }
        anchor_by_ha = {
            'right': 'east',
            'left': 'west',
            'center': ''
            }
        anchor = (f"{anchor_by_va[self.mpl_text.get_va()]} "
                  f"{anchor_by_ha[self.mpl_text.get_ha()]}")
        if anchor == ' ':
            anchor = 'center'
        return anchor

    @cached_property
    def visible(self) -> bool:
        if not self.mpl_text.get_visible():
            return False
        elif self._ax is None or not self.mpl_text.get_clip_on():
            return True
        else:
            return self._is_inside_ax()

    @cached_property
    def _display_xy(self):
        return self.mpl_text.get_transform().transform(
            self.mpl_text.get_position())

    @cached_property
    def _figure_xy(self):
        return self._figure_transform.inverted().transform(self._display_xy)

    @cached_property
    def _axes_xy(self):
        if self._ax is None:
            raise ValueError("This text does not belong to any Axes")
        return self._ax.transAxes.inverted().transform(self._display_xy)

    def _is_inside_ax(self):
        x, y = self._axes_xy
        if (0 <= x <= 1) or (0 <= y <= 1):
            return True
        else:
            return False


@beartype
def extract_text(fig: plt.Figure, /) -> set[FigureText]:
    return remove_transparent(remove_empty(remove_invisible(
        set(get_text_decendents(fig)))))

def remove_invisible(text_set: set, /) -> set:
    return {element for element in text_set if element.visible}

def remove_empty(text_set: set, /) -> set:
    return {element for element in text_set if element.text != ''}

def remove_transparent(text_set: set, /) -> set:
    return {element for element in text_set if element.color[3] != 0}

@beartype
def make_all_transparent(fig: plt.Figure, /) -> dict:
    removed_colors = {}
    for text in get_text_decendents(fig):
        removed_colors[text.mpl_text] = text.color
        text.color = "none"  # avoids messing with the whitespace
    return removed_colors

@beartype
def restore_colors(fig: plt.Figure, colors: dict):
    for text in get_text_decendents(fig):
        text.color = colors[text.mpl_text]

@beartype
def get_text_decendents(fig: plt.Figure, /) -> Iterator[FigureText]:
    stack = [iter(fig.get_children())]
    current_ax = None
    while stack:
        try:
            child = next(stack[-1])
            if isinstance(child, plt.Text):
                yield FigureText(text=child, fig=fig, ax=current_ax)
            else:
                if isinstance(child, plt.Axes):
                    current_ax = child
                stack.append(iter(child.get_children()))
        except StopIteration:
            stack.pop()

@beartype
def get_height_to_width(fig: plt.Figure) -> float:
    width, height = fig.get_size_inches()
    return height/width

def draw_anchors(fig, figure_xy):
    ax = fig.get_children()[1]
    ax.plot(figure_xy[0], figure_xy[1], '+r', clip_on=False,
            transform=fig.transFigure, zorder=20)

def print_family_tree(mpl_object):
    stack = [iter(mpl_object.get_children())]
    print(stack)
    indent = ""
    while stack:
        try:
            child = next(stack[-1])
            print(f"{indent}{child}")
            stack.append(iter(child.get_children()))
            indent = indent[:-2]
            indent += "  |- "
        except StopIteration:
            indent = indent[:-5]
            indent += "- "
            stack.pop()



import matplatex

import matplotlib.pyplot as plt
import numpy as np

def main():
    x = np.linspace(0, 4*np.pi, 300)
    y = 0.1*x*np.sin(x)
    fig, ax = plt.subplots()
    ax.plot(x, y, label="increasing sine wave")
    ax.set_title("Everything is set in the current font and size.")
    ax.set_xlabel("$x$")
    ax.set_ylabel("$f(x)$")
    ax.legend(loc='upper left')
    ax.annotate('$\mathcal{O}$', (0,0), xycoords='figure fraction')
    ax.annotate('$\mathcal{C}$', (0.5,0.5), xycoords='figure fraction')
    ax.annotate('$\mathcal{I}$', (1,1), xycoords='figure fraction')
    ax.annotate('Text colour is kept', (0.4, 0.6), xycoords='figure fraction',
                color='xkcd:pink')
    ax.annotate('So is rotation', (0.5, 0.4), xycoords='figure fraction',
                rotation=20)
    ax.annotate('Label text must be valid \LaTeX code', (0.3, 0.3),
                xycoords='figure fraction', rotation=-2)

    matplatex.save(fig, "./example")
    plt.savefig("example_expected.pdf", format='pdf')
#    plt.show()

if __name__ == '__main__':
    main()



"""matplatex: export matplotlib figures as pdf and text separately for
use in LaTeX.

Copyright (C) 2024 Johannes Sørby Heines

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import sys

from .__about__ import __version__
from .string_replacements import invalid_latex


class LaTeXinput:
    """Text to be input into a LaTeX document to include a figure.

    Instance variables:
    latexcode       The text in question as a string.
    open_graphics   Tracks whether a figure environment is currently
                    open in latexcode.
    widthcommand    The LaTeX length command which will be used to
                    define the width of the figure.

    Public methods:
    __init__            Constructor.
    includegraphics     Open a figure environment and include a figure.
    add_text            Draw a text box.
    endgraphics         Close a figure environment.
    addline             Add a single line of code to latexcode.
    write               Write latexcode to a file.
    """

    # only works for single characters
    # translatex = str.maketrans(invalid_latex)

    def __init__(self, *, widthcommand: str):
        """Constructor for the LaTeXinput class.

        Keyword only arguments:
        widthcommand    The LaTeX length command which will be used to
                        define the width of the figure.
        """
        self.open_graphics = False
        self.widthcommand = as_latex_command(widthcommand)

        self.latexcode = '\n'.join([
            f"% This file was automatically generated by matpLaTeX {__version__}.",
            "% The source code is at https://github.com/johashei/matplatex.",
            "%",
            "% Requires package tikz.",
            "%",
            "% Usage:",
            "%",
            "In the preamble, add",
            rf"%   \newlength{{{self.widthcommand}}}",
            r"%   \newlength{\matplatextmp}",
            "%"
            "% Set the desired with of the figure using",
            rf"%   \setlength{{{self.widthcommand}}}{{<your desired width>}}",
            "%",
            "% Include the figure with",
            rf"%   \input{{<file name>.pdf_tex}}",
            "% or using the import package:",
            rf"%   \import{{<path>}}{{<file name>.pdf_tex}}",
            "%",
            "%"
            ])

    def includegraphics(self, graphics_filename, height_to_width: float):
        """ Start a tikzpicture and include the graphics."""
        if self.open_graphics:
            self.endgraphics()
        self.addline('')
        self.latexcode += '\n'.join([
            r"\begingroup",
            "",
            rf"\setlength{{\matplatextmp}}{{{height_to_width}{self.widthcommand}}}",
            r"\hspace{-\parindent}",
            rf"\begin{{tikzpicture}}[x={self.widthcommand}, y=\matplatextmp]",
            rf"  \node[inner sep=0pt, above right] (graphics) at (0,0) {{",
            rf"    \includegraphics[width={self.widthcommand}]{{{graphics_filename}}}}};",
            ])
        self.open_graphics = True

    def add_text(self, text, position, *,
                 rotation=0, color=(0, 0, 0), alpha=1, anchor='center'):
        if len(color)==4:
            alpha = color[3]
        self.addline(rf"  \node [inner sep=0pt, "
                     rf"text={{rgb,1:red,{color[0]}; "
                                  rf"green,{color[1]}; "
                                  rf"blue,{color[2]}}}, "
                     rf"rotate={rotation}, "
                     rf"anchor={anchor}, "
                     rf"opacity={alpha}] "
                     rf"at ({position[0]}, {position[1]}) "
                     rf"{{{text}}};")

    def endgraphics(self):
        self.addline(r"\end{tikzpicture}")
        self.addline(r"\endgroup")
        self.open_graphics = False

    def addline(self, text):
        self.latexcode += f"\n{text}"

    def write(self, filename):
        if self.open_graphics:
            self.endgraphics()
#        self.addline(rf"\global\let{self.widthcommand}\undefined%")
#        self.latexcode.replace('  ', '\t')
        self.latexcode = replace_multiple(self.latexcode, invalid_latex)
        with open(filename, 'w') as outfile:
            outfile.write(self.latexcode)


def as_latex_command(string: str, /):
    """Make sure the string begins with exactly one backslash."""
    return '\\' + string.strip('\\')

def replace_multiple(string: str, replacements: dict) -> str:
    for key, val in replacements.items():
        string = string.replace(key, val)
    return string

def trim(string):
    """Trim a multiline string.

    Docstring processing algorithm as implemented in PEP 257.
    """
    if not string:
        return ''
    lines = string.expandtabs().splitlines()
    indent = sys.maxsize
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    trimmed = [lines[0].strip()]
    if indent < sys.maxsize:
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)
    return '\n'.join(trimmed)



"""matplatex: export matplotlib figures as pdf and text separately for
use in LaTeX.

Copyright (C) 2024 Johannes Sørby Heines

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from beartype import beartype
import matplotlib.pyplot as plt

from .tools import write_tex, make_all_transparent, restore_colors
from .latex_input import LaTeXinput

@beartype
def save(
        figure: plt.Figure,
        filename: str,
        *,
        widthcommand: str = r"\figurewidth",
        draw_anchors: bool = False,
        verbose: int = 1
        ):
    """Save matplotlib Figure with text in a separate tex file.

    Arguments
    ---------
    figure      The matplotlib Figure to save
    filename    The name to use for the files, without extention

    Optional keyword arguments
    --------------------------
    widthcommand    The LaTeX length command which will be used to
                    define the width of the figure.
    draw_anchors    If True, mark the text anchors on the figure.
                    Useful for debugging.
    verbose: int    0: Print nothing.
                    1: Print save message to stdout. (default)
                    2: Also print runtime info to stderr.
    """
    figure.draw_without_rendering() # Must draw text before it can be extracted.
    output = LaTeXinput(widthcommand=widthcommand)
    write_tex(
        output,
        figure,
        graphics=filename,
        add_anchors=draw_anchors,
        verbose=(verbose==2)
        )
    output.write(f"{filename}.pdf_tex")
    color_backup = make_all_transparent(figure)
    figure.savefig(f"{filename}.pdf", format='pdf')
    restore_colors(figure, color_backup)
    if verbose:
        print(f"Figure written to files {filename}.pdf_tex and {filename}.pdf")



from pathlib import Path
from subprocess import run

import matplotlib.pyplot as plt
import pytest

from matplatex import save

LATEX_TYPICAL = r"""
\documentclass{article}

\usepackage{graphicx}
\usepackage{tikz}

\newlength{\figurewidth}
\newlength{\matplatextmp}

\begin{document}

\setlength{\figurewidth}{\linewidth}
\input{figure.pdf_tex}

\end{document}
"""

@pytest.fixture
def figure():
    plt.rc('text', usetex=True)
    fig, [ax1, ax2] = plt.subplots(1, 2)
    x1 = [-1, -0.5, -0.3, -0.1, 0.1, 0.3, 0.5, 1]
    y11 = [x**3 for x in x1]  # no numpy to avoid superfluous dependencies
    y12 = [x**2 for x in x1]
    ax1.plot(x1, y11, 'o-', label='$x^3$')
    ax1.plot(x1, y12, 'd-', label='$x^2$')
    ax1.legend()
    ax1.set_xlabel('x', usetex=False)
    ax1.set_ylabel('y')
    ax2.axhline(20)
    ax2.axhspan(0, 17)
    ax2.plot([0, 3, 4, 7], [15, 11, 7, 3], '--')
    return fig

def test_compilation(figure, tmp_path):
    latex_path = tmp_path / 'document.tex'
    figure_path = tmp_path / 'figure'
    # Regenerate the files each time so changes are applied.
    latex_path.write_text(LATEX_TYPICAL, encoding='utf-8')
    save(figure, str(figure_path), verbose=2)
    # compile with pdflatex
    compilation = run(
        ['pdflatex', latex_path.name],
        cwd=tmp_path
        )
    assert compilation.returncode == 0



import matplotlib.pyplot as plt
from matplotlib.colors import to_rgba
import pytest

import matplatex.tools as tools


@pytest.fixture
def make_simple_figure():
    """Make a figure with a single subplot and unique text objects."""
    fig, ax = plt.subplots()
    ax.plot([0,2], [1,0])
    ax.set_xlabel("x axis label")
    ax.set_ylabel("y axis label")
    ax.set_title("axis title")
    ax.set_xticks([0, 1, 2], ['none', 'I', 'II'])
    ax.set_yticks([0, 1/3, 2/3, 1], ['0', '1/3', '2/3', '1'])
    visible_text_objects = set(
        [ax.title, ax.xaxis.label, ax.yaxis.label]
        + ax.get_xticklabels()
        + ax.get_yticklabels()
        )
    return fig, visible_text_objects

def test_get_text(make_simple_figure):
    result = {text.mpl_text for text in
              tools.extract_text(make_simple_figure[0])}
    expected = make_simple_figure[1]
    assert result == expected

def test_make_all_transparent(make_simple_figure):
    tools.make_all_transparent(make_simple_figure[0])
    assert tools.extract_text(make_simple_figure[0]) == set()

def test_restore_colors(make_simple_figure):
    initial_state = {t.mpl_text for t in
                     tools.extract_text(make_simple_figure[0])}
    removed_colors = tools.make_all_transparent(make_simple_figure[0])
    tools.restore_colors(make_simple_figure[0], removed_colors)
    final_state = {t.mpl_text for t in
                   tools.extract_text(make_simple_figure[0])}
    assert initial_state == final_state

@pytest.fixture
def text_only_figure():
    fig = plt.Figure(figsize=(10, 6))
    fig.add_artist(plt.Text(0.42, 0.2845, 'text', color='goldenrod'))
    return fig

def test_get_position_in_figure(text_only_figure):
    text = tools.extract_text(text_only_figure)
    result = text.pop().position_in_figure
    expected = (0.42, 0.2845)
    tolerance = 1e-4
    assert abs(result[0] - expected[0]) < tolerance
    assert abs(result[1] - expected[1]) < tolerance

def test_get_color(text_only_figure):
    text = tools.extract_text(text_only_figure)
    result = text.pop().color
    expected = to_rgba('goldenrod')
    assert result == expected

def test_get_height_to_width(text_only_figure):
    result = tools.get_height_to_width(text_only_figure)
    expected = 6/10
    assert result == expected

@pytest.fixture
def figure_with_clipped_text():
    fig, ax = plt.subplots()
    ax.set_xticks([])
    ax.set_yticks([])
    ax.add_artist(plt.Text( 0.7, 0.2, 'text inside axes'))
    ax.add_artist(plt.Text( 1.1, 1.8, 'text outside axes'))
    return fig

def test_is_inside_ax(figure_with_clipped_text):
    expected_by_text = {
        'text inside axes': True,
        'text outside axes': False}
    for text in tools.extract_text(figure_with_clipped_text):
        result = text._is_inside_ax()
        expected = expected_by_text[text.text]
        assert result == expected

@pytest.fixture
def figure_with_multiple_axes():
    fig, [ax1, ax2] = plt.subplots(1,2)
    fig.add_artist(plt.Text(0.5, 0.5, 'figure text'))
    ax1.annotate('ax1 text', (3.14, 6.28))
    ax2.annotate('ax2 text', (0.35, 6.6))
    return fig

def test_ax_identification(figure_with_multiple_axes):
    expected_by_text = {
        'figure text': None,
        'ax1 text': figure_with_multiple_axes.get_axes()[0],
        'ax2 text': figure_with_multiple_axes.get_axes()[1]}
    for text in tools.extract_text(figure_with_multiple_axes):
        result = text._ax
        expected = expected_by_text[text.text]
        assert result == expected
